from numpy import *
import sys
import globalvars80 as g
import scipy.optimize as sciop
import rout80 as rout
import OK10
import time

def eh_eqan (Rc, s, sig, eh_pv, Tfr_pv):
  """
  solves for the Hill eccentricity in the equilibrium regime,

    Tstir (eh) = Tdamp (eh)

  utilizing the power-law prescription for Tvs(eh)

  Tdamp is the minimum of the friction time with the gas and, if
  implemented, the decay timescale due to bouning collisions.

  Tfr_pv and eh_pv are previous solutions (guesses) for the eccentricity
  and friction time

  Ouput: eh and the corresponding Tf and Tdamp
  """
  #Timescales corresponding to transitions in P_vs
  Tvs1 = T_vs (g.eh1, Rc)
  Tvs2 = T_vs (ehesc, Rc)

  #and now start the iteration
  eh = eh_pv
  Tfr = Tfr_pv
  iiter = 0
  err = 1.0
  while abs(err) >g.tol:
    iiter += 1
    eh_p = eh
    Tfr_p = Tfr
    [Tdamp, Tfr] = T_damp(eh, sig, s, Tfr_p)
    eh = sol_eh (Rc, Tdamp, Tvs1, Tvs2)#this is much quicker than brentq
    err = abs(eh-eh_p)/eh
    if iiter>50:
      if g.vbl>=0: print "[eh_eqsol]Warning: >50 iter needed with error at", err, s
      break

  Tstir = Tdamp
  Tgrow = T_grow (eh, Rc, sig, s, Tfr)
  return [eh, Tfr, Tdamp, Tstir, Tgrow]


def T_damp (eh, sig, s, Tfrpv):
  """
  Returns the damping timescale

  We compute the collision timescale among bodies, which is relevant when
  collisional cooling (bouncing is implemented).  Here we assume that i=e/2 for
  simplicity calculate the friction timescale and set the limiting timescale
  (Tdamp)
  """
  Tfr = T_fric (eh, s, Tfrpv)
  if g.doBO: 
    Tcol = T_col (s, sig, eh, 0.5*eh*hM)
    Tdamp = min(Tfr, Tcol/(1-g.cofr))
  else: 
    Tdamp = Tfr
  return [Tdamp, Tfr]


def eh_eqsol (eh, s, sig, Tfrpv):
  """
  Obtains eh from Tstir == Tdamp
  """
  [Tdamp, Tfr] = T_damp(eh, sig, s, Tfrpv)
  Tstir = T_stir (eh)
  return Tdamp-Tstir


def eh_eq (Rc, s, sig, eh_pv, Tf_pv):
  """
  solve for eh using scipy's brentq (slow).  
  Alternatively, EH_EQAN can be used
  """
  ehmin = rout.eh_min(s,vh) #minimum eh (from self-stirring plts)

  eh = sciop.brentq(eh_eqsol, ehmin, ehesc, args=(s, sig, Tf_pv), xtol=g.tol)

  [Tdamp, Tfr] = T_damp(eh, sig, s, Tf_pv)
  Tstir = Tdamp
  Tgrow = T_grow (eh, Rc, sig, s, Tfr)
  return [eh, Tfr, Tdamp, Tstir, Tgrow]


def eh_frsol (eh, s, sig, Tfr_pv):
  """
  Obtains eh from Tgr == Tdamp
  """
  [Tdamp, Tfr] = T_damp(eh, sig, s, Tfr_pv)
  Tgrow = T_grow (eh, Rc, sig, s, Tfr)

  return Tgrow-Tdamp


def eh_fr (s, sig, eh_pv, Tfr_pv):
  """
  Obtain eh in the friction time regime, Tgrow=Tstir
  """
  ehmin = rout.eh_min(s,vh) #minimum eh (from self-stirring plts)

  if eh_frsol (ehmin, s, sig, Tfr_pv)>0:
    #return smallest eccentricity
    eh = ehmin
  elif eh_frsol(ehesc, s, sig, Tfr_pv)<0:
    eh = ehesc
  else:
    eh = sciop.brentq(eh_frsol, ehmin, ehesc, args=(s, sig, Tfr_pv), xtol=g.tol)

  [Tdamp, Tfr] = T_damp(eh, sig, s, Tfr_pv)
  Tgrow = T_grow (eh, Rc, sig, s, Tfr)
  Tstir = T_stir (eh)
  return [eh, Tfr, Tdamp, Tstir, Tgrow]


def eh_nesol (eh, Rc, s, sig, Tfr_pv):
  """
  Obtains eh from Tgr == Tstir

  Note that we assume friction times are unimportant in calculation the
  eccentricity in this regime and are put infinity (which decreases the
  computation time)
  """
  Tstir = T_stir (eh)
  Tgrow = T_grow (eh, Rc, sig, s, inf)

  return Tgrow-Tstir


def eh_ne (Rc, s, sig, eh_pv, Tfr_pv):
  """
  Obtain eh in the non-equilibrium regime, Tgr=Tstir
  """
  ehmin = rout.eh_min(s,vh) #minimum eh (from self-stirring plts)

  if eh_nesol(ehmin, Rc, s, sig, Tfr_pv)<0:
    #return smallest eccentricity
    eh = ehmin
  elif eh_nesol(ehesc, Rc, s, sig, Tfr_pv)>0:
    eh = ehesc
  else:
    eh = sciop.brentq(eh_nesol, ehmin, ehesc, args=(Rc, s, sig, Tfr_pv), xtol=g.tol)

  [Tdamp, Tfr] = T_damp(eh, sig, s, Tfr_pv)
  Tstir = T_stir (eh)
  Tgrow = T_grow (eh, Rc, sig, s, inf)
  return [eh, Tfr, Tdamp, Tstir, Tgrow]


def P_col(eh, al_a, Rh, inc, tau=-1):
  """
  Returns the dimensionless accretion rate Pcol

  Due to atmosphere enhancemnt the capture radius (R_a) will in generally
  become larger than the core radius R_c.  Thus, al_a >= al, in dimensionless
  (Hill) units.

  If a non-infinite tau number is given AND the doND flag is set, we instead
  use the drag-modified expression of Ormel & Klahr (2010) to calculate Pcol
  """

  #Pcol in the 2D regime (razor-thin disk)
  Pcol2d = 11.3*al_a**0.5
  if eh>g.ehsd:#d.d.-regime
    #The latter corrects for a different inclination 
    #model (ie Pollack)
    ih = inc/hM
    Pcol = g.Ccol*al_a/eh**2 *(0.5*eh/ih)
    if Pcol>Pcol2d: Pcol = Pcol2d
  else:#s.d. regime
    if g.doND and tau<inf:#apply Ormel and Klahr (2010)
      zeta_w = g.vhw/vh
      [Pcol, bcol] = OK10.P_col(tau, zeta_w, al_a)
      bcol = bcol*Rh#includes Hill radius
    else:
      Pcol = Pcol2d
      bcol = 1.7 *al_a**0.5 *Rh
    #Pcol is now corrected by the scaleheight
    heff = rout.h_eff(tau, inc, g.cs/g.Om0)
    Pcol = Pcol*min(1,bcol/heff)

  return Pcol


def sig_New (sig, Tstep, Ts, sig_ss):
  """
  Solves the differential equation

    dsig/dt = (-sig +sig_ss)/Ts

  sig_ss  :sigma in the steady-state limit
  Ts      :corresponding timescale.

  Use the well-known solution of the differential equation dy/dx = -y +y_s:

    y = ys +(y0-ys)*exp(-x)
  """
  arg = Tstep/Ts#this is in the above

  if arg>1e-10:
    sigNew = sig_ss -(sig_ss-sig)*exp(-arg)
  else:#to counter near-cancellation
    sigNew = sig +(sig_ss-sig)*arg

  return sigNew


def T_col (sp,sig,eh,inc):
  """
  Obtain the collision timescale

  Note: we neglect the scaleheight increase due to diffusion here
  """
  ecc = eh*hM #orbital eccentricity
  if sig==0: Tcol = inf
  else: Tcol = (1./3) *sp*g.rhos/(sig*g.Om0) *(2*inc/ecc)
  return Tcol


def T_drift(sp,Tfr,Rh,vran):
  """
  Calculates the drift timescale, T_drift

  When there is no drift, doRD==0 it is by definition inifinite

  Otherwise, it is given by Wsim/vdrift, where Wsim is a characteristic length.
  The latter can be choosen either to represent the feeding zone of the
  protoplanet (~Rh) or will be similar to the semi-major axis (~a0). Cdrift is
  a fudge parameter

  The former corresponds to a single protoplanet (doOL==0), while in the latter
  we assume the presence of an oligarchy (doOL==1), where the evolution of the
  neighboring oligarchs proceeds similar.
  """
  if g.doRD==False:
    Tdrift = inf
  else:
    tau = Tfr*g.Om0
    #The drift timescale scales with ~semi-major axis to account 
    #for particles at exterior orbits drifting in 
    #Cdriftfac sets the (effective) length scale
    vdrift = rout.v_drift (tau, vran)
    if g.doOL: Wsim = g.Cdriftfac *g.a0
    else: Wsim = g.btwig *Rh
    Tdrift = Wsim/vdrift

  return Tdrift


def T_growsol (Tgr, Rc, eh, inc, sig, s, sa, Tfr):
  """
  This solves the implicit equation T_grow(eh,Tgr) = Tgr.  Note that the growth
  timescale depends on itself (Tgr), when atmosphere enhancement (AE) is
  implemented.  That is, the structure of the atmosphere depends on the rate at
  which it captures solids, i.e., the growth timescale.
  """
  Tgr_nw = T_grownew (Rc,eh,inc,sig,s,sa,Tfr,Tgr)
  return Tgr_nw-Tgr


def T_grownew (Rc, eh, inc, sig, s, sa, Tfr, Tgrow, doAtmos=True):
  """
  Compute the growth timescale, optionally with inclusion of:
    - the atmosphere enhancement (AE)
    - drag enhancement [Ormel & Klahr (2010) collision rates] (ND)
    - turbulent diffusion (TD)

  The function calls R_a to find the (enhanced) capture rate due AE.  This
  becomes the new "size" (i.e., the capture radius), and is used for P_col
  """
  if doAtmos:
    [Ra, Wneb] = rout.R_a (Rc, Rh, Rb, Mc, Tgrow, eh, sa, g.Tneb, rhog, g.gamAd)#the enhancement radius
    al_a = Ra/Rh#the new radius
  else:
    al_a = Rc/Rh

  #Next, we compute the dimensionless Pcol
  Pcol = P_col(eh, al_a, Rh, inc, tau=Tfr*g.Om0)
  dMdt = Cacc*Pcol*vh*Rh*sig#conversion to physical units
  Tgrownew = Mc/dMdt

  return Tgrownew


def T_grow (eh, Rc, sig, s, Tfr):
  """
  solves for the growth timescale recursively.  The solution is implicit since
  the growth timescale depends on itself when the doAE is set.  The function
  T_growsol solves the implicit equation:

    Tgrow = T_grow(Tgrow)
  """
  #Obtain the orbital inclination
  inc = rout.inc_val (eh*hM, s)
  Tgrow=-1
  if sig==0:
    Tgrow = inf
  else:
    #The growth timescale without atmospheres
    Tgrow_na = T_grownew (Rc, eh, inc, sig, s, -1, Tfr, Tgrow, doAtmos=False)
    if g.doAE and Tfr<>inf:
      #the required overdensity
      rho_a = (6+eh**2)*s*g.rhos /(9*Rh)
      sa = rho_a/rhog
      #!TBD: the lower limit for Tgrow is arbitrary (although prob. OK)
      Tgrow = sciop.brentq (T_growsol, Tgrow_na/1e3, 1.001*Tgrow_na,\
          args=(Rc,eh,inc,sig,s,sa,Tfr,), xtol=g.tol*Tgrow_na)
      #Now we must call again to obtain Ra and Wneb
      [Ra, Wneb] = rout.R_a (Rc, Rh, Rb, Mc, Tgrow, eh, sa, g.Tneb, rhog, g.gamAd)
      if Wneb>1.0 and g.vbl>=0: 
        print 'Warning[T_grownew]: Wneb>1(',Wneb, '); atmosphere approximation formally invalid'
    else:
      Tgrow = Tgrow_na
  return Tgrow


def T_sigF (sigF, sigE, TgrF, TdriftF, sigdotPF, sigIN=0):
  """
  The timescale for the depletion of the fragments' surface dentity.  We
  account for increase/decrease of fragments due to:
    -planetesimal fragmentation (+)
    -embryo growth (-)
    -fragment drift out of feeding zone (-)
    -(optional) fragment external influx (+)

  TsF collects all the terms linear in surface density; that is:
  
     dsigF/dt = -sigF/TsF + sigFss/TsF
  
  where sigFss is the steady-state sigma
  """
  #compute the rate without the drift term and the total rate
  sigdotFE = -sigE/TgrF/Cacc 
  sigdotF = sigdotPF +sigdotFE +sigIN -sigF/TdriftF

  if sigF==0: TsF=inf
  else: TsF = (1/TdriftF -sigdotFE/sigF)**-1

  sigF_ss = (sigdotPF +sigIN) *TsF

  if sigF==0: 
    TsigF = inf
  else:#analytically solve for the right timescale 
    TsigF = sigF /sigdotF

  #sigdotF is the rate w/o the drift term and is only used when sigF=0
  return [TsigF, sigdotF, sigF_ss, TsF]


def T_sigP (sigP, sigE, TgrP, TdriftP, sigdotPF, sigIN=0):
  """
  The timescale for the change in the planetesimals' surface dentity
  distribution.  The timescale is influenced due to:
    -planetesimal drift (-)
    -planetesimal fragmentation (-)
    -embryo growth (-)
    -(optional) planetesimal influx

  It returns the rate and the timescale
  """

  sigdotP = -sigP/TdriftP -sigdotPF -sigE/TgrP/Cacc +sigIN
  if sigP==0: TsigP = inf
  else: 
    TsigP = sigP/sigdotP
  return [sigdotP, TsigP]


def T_vs (eh, Rc, Tzero=0):
  """
  Returns the viscous stirring timescale as function of eccentricity and embryo
  size.
  """
  Pvs = rout.P_vs(eh)
  NcolE = (2*pi*g.a0 *Rh*g.btwig)**-1
  Tvs = 2*eh**2 /(Pvs*Rh*vh*NcolE)

  return Tvs-Tzero


def T_ts (eh, Tzero=0):
  """
  Returns the turbulent stirring timescale, Tts.
  """
  Tts = preTts *(eh*hM)**2
  return Tts


def T_stir (eh):
  """
  Returns the stirring timescale, including, optionally turbulent stirring
  """
  Tvs = T_vs (eh, Rc, Tzero=0)
  if g.doTS:
    Tts = T_ts (eh, Tzero=0)
    Tstir = min(Tts, Tvs)
  else:
    Tstir = Tvs

  return Tstir


def sol_eh (Rc, Tvs, Tvs1, Tvs2):
  #COM: that i=e/2 is assumed.
  #TBD0.7-2: more general stirring may be considered
  """
  Analytically solves for eh, for a given Tvs. 

  When TS (turbulent stirring) is included, take the largest eh

  Tvs1, Tvs2 are the break points in our segmented T_vs curve.  These
  correspond to eh1 (2.34) and ehesc (escape velocity), respectively
  """
  if Tvs<=Tvs1:#for eh<eh1
    eh = g.eh1*sqrt(Tvs/Tvs1)
  elif Tvs>=Tvs2:#for eh>eh_esc
    eh = ehesc#saturates out at escape velocity
  else: 
    eh = g.eh1*(Tvs/Tvs1)**0.25

  if g.doTS:
    ehTS = sqrt(Tvs/preTts) /hM 
    if ehTS>eh: 
      eh=ehTS

  return eh


def T_fric (eh, s, Tfr_pv):
  """
  Iteratively compute the friction time as function of Hill eccentricity.

  Since the friction time determines the relative velocity with the gas (vgas),
  hence the iteration.
  """
  Tfr = Tfr_pv
  err = 1.0
  iiter = 0
  while abs(err)>g.tol:
    iiter += 1
    vgas = rout.calc_vgas (Tfr, eh*vh, s)
    Tfr_p = Tfr
    Tfr = rout.calc_Tau (s, g.cs, rhog, vgas)
    err = abs(Tfr -Tfr_p) /Tfr
    if iiter>50:
      if g.vbl>=0: print "[T_fric]Warning: >50 iter needed with error at", err
      break

  return Tfr


def try_mode (Rc, s, eh_pv, sig, Tfr_pv, mode):
  """
  implements the conditions for MODE where MODE is one of:
    
    mode = ['eq', 'ne', 'fr']

  corresponding to the three regimes

  *eh_eqan* custom-solves for eh, which is faster (but a bit tricker) than
  using the brentq solver using eh_eq

  con gives the condition regarding the regime.  This should be true, of course
  """
  if mode=='eq':
    [eh, Tfr, Tdamp, Tstir, Tgrow] = eh_eqan (Rc, s, sig, eh_pv, Tfr_pv)
    con = Tdamp<Tgrow
  elif mode=='ne':
    [eh, Tfr, Tdamp, Tstir, Tgrow] = eh_ne (Rc, s, sig, eh_pv, Tfr_pv)
    con = Tgrow<Tdamp
  elif mode=='fr':
    [eh, Tfr, Tdamp, Tstir, Tgrow] = eh_fr (s, sig, eh_pv, Tfr_pv)
    con = Tgrow<Tstir
  else:
    print "Error[try_mode]: No valid growth mode!"
    stop

  return [eh, Tfr, Tdamp, Tstir, Tgrow, con]


def calc_Gmodes (Rc, s, sig, inArr_pv, force=''):
  """
  This routine calculates the growth-modes (equilibrium, non-eq) and assigns
  the appropriate one

  First it tries the previous regime as given by MODE_PV.  If the condition
  fails we search for another regime.  In that case we choose the new regime
  based on the proximity to the previous eccentricity

  A growth regime can be forced by setting the force to one in ['eq', 'ne',
  fr']
  """
  [mode_pv, eh_pv, Tgr_pv, Tfr_pv] = inArr_pv
  modes = ['ne', 'fr', 'eq']
  errs = ones((3),dtype=float)
  imode0 = modes.index(mode_pv)
  imode = imode0
  mode = modes[imode]
  if force=='':
    [eh, Tfr, Tdamp, Tstir, Tgrow, con] =\
        try_mode (Rc, s, eh_pv, sig, Tfr_pv, mode)
    while con==False:
      if imode>imode0+1:
        imode = argmin(errs)#this is the closest
        con = True
      else:
        imode += 1
      mode = modes[mod(imode,3)]
      [eh, Tfr, Tdamp, Tstir, Tgrow, con2] =\
          try_mode (Rc, s, eh_pv, sig, Tfr_pv, mode)
      if eh_pv<>-1 and con2:#These are conditions
        err = abs(eh-eh_pv)/(eh+eh_pv)
        errs[mod(imode,3)] = err
  else:
    [eh, Tfr, Tdamp, Tstir, Tgrow, con] =\
        try_mode (Rc, s, eh_pv, sig, Tfr_pv, force)

  return [mode, eh, Tgrow, Tfr]


def Mdot_ext(time):
  return 0.0

#now we really start
def main (Time=g.Tini, fmode='', fn_outputtable='data.dat', Mdot_ext=Mdot_ext):
  t_start = time.clock()
  #start with definition of the global variables

  #Disk parameters are constant throughout
  #These are either obtained by calling the mmsn() function or from an
  #input file that is assumed to be 'physpar.txt' by default
  if g.useMmsn:
    [g.SigGini, g.SigTini, g.rhog0, g.Tneb]\
        = array(rout.mmsn(['Sig', 'Sig_d','rhog', 'Temp'], a=g.a0/g.au, fsigd=g.fsigd, fsig=g.fsig, M_star=g.Mstar)) 
  else:
    parDict = rout.load_matPars(fname=g.physparName)
    [g.SigGini, g.SigTini, g.rhog0, g.Tneb] = rout.assign_matProps(parDict, ['Sig', 'SigTini','rhog', 'Temp'])
  #Other non-independent disk parameters
  g.Om0 = rout.get_Omega (g.a0, g.Mstar)
  g.cs  = rout.get_cs (g.Tneb)

  #Auxiliary parameters for solving eccentricity values
  global preTts, Cacc, ehesc
  ehesc = rout.eh_max (g.a0, g.rhoc, g.Mstar)

  if g.doOL: Cacc=1.5 
  else: Cacc=1.0
 #preTts = 5e-3/g.alSS *(g.cs**2 *g.SigGini/g.Mstar/g.Om0**2)**-2 *g.Om0**-1
  preTts = rout.get_preTts (g.alSS, g.a0, g.cs/g.Om0, g.SigGini, g.Mstar, g.Om0)

  #Load the material parameters from the material file
  parDict = rout.load_matPars(fname=g.coregrowthdir+'/doc/'+g.mparfile)
  [g.Q0s, g.Q0g, g.bs, g.bg, g.Cgg] = rout.assign_matProps(parDict, ['Q0s', 'Q0g', 'bs', 'bg', 'Cgg'])

  #non-constant global variables
  global al, Rh, Mc, vh, Rc, hM, Rb, rhog, irun


  #Load the dictionary with key model parameters
  #these may be used as the output
  variablesDict = rout.loadvariables()

  #Print model parameters
  if g.vbl>=2:
    print '---------'+' Variables [cgs] '+'---------'
    rout.print_iniPars()
    rout.print_iniPars(['SigGini', 'SigTini', 'rhog0', 'cs'], [g.SigGini, g.SigTini, g.rhog0, g.cs])
    print '---------------------------------'

  #Initialization: embryo properties (Mc, Rh); initial planetesimal size
  #(sP); and surface densities (sigX)
  Mc = g.Mini
  Rh = rout.R_hill(g.a0, Mc) 
  sigEini = rout.sigma_E(Mc,Rh,g.a0,g.btwig)#initial sigma in embryos
  sP = g.sPini
  if g.sFini==-1: sF = rout.sF_ini()
  else: sF = g.sFini
  sigPini = (g.SigTini -sigEini) *g.fsigP
  sigP = sigPini
  sigF = g.SigTini -sigEini -sigPini
  sigE = sigEini
  sigdotF = 0

  #Initialization of variables
  #TBD check if we can put these numbers to infinity, 0, or -1
  #and whether we should really assign these
  #(I think we should, in fact)
  TsigF = inf#TBD: rename to TsigF
  TdriftF = inf
  Tgrow = inf
  Tstep = inf
  TfrF = -1
  qF=-1
  ehF = -1

  modeP = 'ne'
  modeF = 'eq'
  ehP = -1
  TgrP = -1
  TgrF = inf
  TfrP = -1
  TfrF = -1

  irun=0#number of iterations
  #Start of the main loop
  while (Time<g.Tend and Mc<1.1e2*g.Mearth and (sigP+sigF)>g.fsigmin*g.SigTini):

    #the previous state
    inArrP_pv = [modeP, ehP, TgrP, TfrP]
    inArrF_pv = [modeF, ehF, TgrF, TfrF]

    #update the core values (radii, etc.)
    Rb = rout.R_bondi(Mc, g.cs, g.gamAd)
    Rc = rout.mass2rad(Mc, g.rhoc)
    Rh = rout.R_hill(g.a0, Mc) 
    vh = Rh*g.Om0
    hM = Rh/g.a0
    sigE = rout.sigma_E(Mc,Rh,g.a0,g.btwig)#current surface density in embryos
    rhog = g.rhog0 *exp(-Time/g.Tdisp)

    #This computes the dynamical state (provides the eccentricity)
    if sigP>0:
      [modeP, ehP, TgrP, TfrP] = calc_Gmodes (Rc, sP, sigP, inArrP_pv, force=fmode)
      #Material strength and qP value
      QdP = rout.Q_d(sP)
      qP = 0.125*(ehP*vh)**2 /QdP
     #print irun, modeP, ehP, qP, sigP, Mc/g.Mearth

    #Obtain the inclinations and scaleheights of the particles
    incP = rout.inc_val(ehP*hM, sP)
    incF = rout.inc_val(ehF*hM, sF)
    hP = rout.h_eff(TfrP*g.Om0, incP, g.cs/g.Om0)
    hF = rout.h_eff(TfrF*g.Om0, incF, g.cs/g.Om0)

    #collision and drift timescales among plts
    TcolP = T_col (sP, sigP, ehP, incP)
    TdriftP = T_drift(sP, TfrP, Rh, ehP*vh)

    #Fragment properties
    if sigF>0:
      [modeF, ehF, TgrF, TfrF] = calc_Gmodes(Rc, sF, sigF, inArrF_pv, force=fmode)
      qF = 0.125*(vh*ehF)**2 /rout.Q_d(sF)
      TcolF = T_col (sF, sigF, ehF, incF)

    #Total growth timescale
    Tgrow = (1/TgrP + 1/TgrF)**-1


    #Calculation of surface density inflow
    if g.doOL==False: 
      sigdotFext = Mdot_ext(Time) /(2*pi*g.a0) /(g.btwig*Rh) \
                   +(g.SigTini-sigF)/(3*Tgrow)
      sigdotPext = 0
    else: 
      sigdotFext = 0
      sigdotPext = 0

    #Calculate surface density depletion timescale (fragments)
    if g.doFR and sigP>0:
      sigdotPF = -rout.dsigP_dt (sigP,sigF,qP,TcolP,hP,hF)
    else:
      sigdotPF = 0

    if g.doFR: 
      TdriftF = T_drift(sF, TfrF, Rh, ehF*vh)
      [TsigF, sigdotF, sigF_ss, TsF] =\
          T_sigF (sigF, sigE, TgrF, TdriftF, sigdotPF, sigIN=sigdotFext)

    #Calculate surface density depletion timescale (planetesimals)
    [sigdotP, TsigP] = T_sigP (sigP, sigE, TgrP, TdriftP, sigdotPF, sigIN=sigdotPext)

    #variables to write out are given in sQarr
    #TBD0.7-4: the eval() function is slow and apparently dangerous; 
    #look for alternative?
    Qarr = [eval(s) for s in g.sQarr]
    rout.writeData(irun, g.sQarr, Qarr, fname=fn_outputtable)

    #increase iteration count
    irun += 1

    #Construct the array of key timescales
    #and pick the minimum
    Tstepold = Tstep
    if g.doPG==g.doFR==False: 
      #nothing happens with the planetesimals so remove TcolP from the list
      Tarr = [TgrP, TdriftP, abs(TsigP), TgrF, TsigF]
    else: 
      Tarr = [TgrP, TdriftP, TcolP, abs(TsigP), TgrF, TsigF]
      Tarr = [TgrP, TsigP, TgrF, TsigF, TcolP]
    imin = argmin(abs(array(Tarr)))
    #Tstep = min(g.xi*abs(Tarr[imin]), (1+g.xi)*Tstepold)
    Tstep = g.xi*abs(Tarr[imin])

    if g.vbl>=3: print irun, imin, Tstep/g.yr/1e3, TsigF/g.yr/1e3, sigF

    #changes of characteristic size planetesimals (sP)
    #if g.doPG and qP<g.Qtc:
    if g.doPG:
      sP = sP*exp(Tstep/(3*TcolP))

    #changes of characteristic size fragments (sF)
    #TBD0.7-5: this is a rather brute force and somewhat time-consuming way
    #of obtaining sF. Improve!
    while sigF>0 and sF>g.s_min and qF>g.qFfrag:
      sF = sF/1.01
      [modeF, ehF, TgrF, TfrF] = calc_Gmodes(Rc, sF, sigF, inArrF_pv, force='eq')
      qF = 0.125*(vh*ehF)**2 /rout.Q_d(sF)


    #Update surface density planetesimals (sigP)
    if sigP==0: sigP = sigdotP*Tstep
    elif sigP<g.SigTini*g.fsigmin: sigP=0
    else: sigP = sigP *(1 + Tstep/TsigP)

    #Update surface density fragments (sigF)
    #A bit more complicated due to the quasi-steady state that may
    #emerge
    if g.doFR:
      if sigF==0:#the 1st fragmentation event
        sigFnew = sigdotF*Tstep
      else:
        sigFnew = sig_New (sigF, Tstep, TsF, sigF_ss)
      #put zero when minimum is crossed
      if sigFnew<g.SigTini*g.fsigmin: sigF=0
      else: sigF = sigFnew

    #Update embryo mass and time
    Time += Tstep
    Mc *= (1 + Tstep/Tgrow)

  t_end = time.clock()
  if g.vbl>=1: 
    ctime = t_end-t_start
    stime = '%.2f'%(1e3*ctime/irun)
    print irun, "iterations needed in ", ctime, "s  ;",  stime, "ms/iter"
  Qarr = [eval(s) for s in g.sQarr]
  rout.writeData(irun, g.sQarr, Qarr, fname=fn_outputtable)

  return irun

