from numpy import *
import sys
import string
import globalvars80 as g

def C_D(Re_p):
  """
  Returns the drag coefficient given the Reynolds particle number
  This follows Weidenschilling (1977)
  """
  if Re_p<1: Cd = 24/Re_p
  elif Re_p<800: Cd = 24*Re_p**-0.6
  else: Cd = 0.44
  return Cd


def M_iso(a0, sig, dela0, Mstar):
  """
  returns the isolation mass

  dela0 (the width) is expressed in terms of Hill radii of the planet
  """
  Miso = (2*pi*dela0)**1.5 /3**0.5 *a0**3 *sig**1.5 /Mstar**0.5
  return Miso


def P_vs (eh, Pvs0=0):
  """
  Returns the dimensionless viscous stirring rate according to the (simplified)
  fit to the Ohtsuki et al. (2002) curve.
  """
  out = min(g.Pvs2d, g.Pvs2d*(g.eh1/eh)**2) 
  return out-Pvs0


def Q_d (s):
  """
  Obtain the material strength.  Formula contains contributions from:
    -the strength regime (Q0s, bs)
    -the gravity regime (Q0g, bg)
    -the grav. binding energy (Cgg)
  """
  mass = rad2mass(s, g.rhos)
  return  g.Q0s   *s**g.bs\
         +g.Q0g   *g.rhos *s**g.bg\
         +g.Cgg*2 *g.Gn*mass/s


def R_a (Rc, Rh, Rb, Mc, Tgrow, eh, sa, Temp, rhog, gamfac):
  """
  computes the radius enhancement in the presence of atmospheres
  (i.e., the capture radius)
  """

  #key quantities
  #Wneb = W_neb (Rc, Tgrow, Temp, rhog)
  Wneb = g.kB/(16*g.sB*g.mu) *g.kappa*rhog*g.rhoc/Temp**3 *Rc**2/Tgrow
  s1 = max(1, 1/(5*Wneb))

  #distinghuis which regime applies
  if sa<=1: 
    xa = 1#(formally nebinite, but cap at Rc)
  elif sa<s1: 
    xa = 1/(1 +(2*Wneb*(sa-1) +log(sa))/gamfac)
  else: 
    x1 = 1/(1 +(2*Wneb*(s1-1) +log(s1))/gamfac)
    xa = (1/x1 +4/gamfac*(4*Wneb)**(1./3) *(sa**(1./3) -s1**(1./3)))**-1

  Ra = xa*Rb#the enhanced radius
  if Ra<Rc: Ra=Rc
  elif Ra>min(Rb,Rh): Ra=min(Rb,Rh)

  return [Ra, Wneb]


def R_bondi (Mass, cs, gamfac=1.4):
  """
    Provides the Bondi Radius, r_b=GM/c**2
    where cs is the isothermal sound speed

    The adiabatic correction (gamfac) is applied
  """
  return g.Gn*Mass/(gamfac*cs**2)


def R_hill (a0, mass):
  """
    Returns the Hill radius
  """
  return a0*(mass/3/g.Mstar)**(1./3)


def Re_p (vgas, nu_mol, s):
  return 2*s*vgas/nu_mol


def alpha_a (a0, rho=1.0, Mstar=g.Msun):
  """
    computes alpha = R/R_h as function of disk radius

    Input parameters:
      - a0    = disk radius
      - rho   = density of the body
      - Mstar = mass of star
  """
  al = (9*Mstar /(4*pi*rho))**(1./3) /a0
  return al


def assign_matProps (parDict, skeys):
  Qarr = []
  for s in skeys:
    Qdum = float(parDict[s])
    Qarr.append(Qdum)
  return Qarr


def calc_vgas (Tfr, vran, s):
  #laminar velocity
  St = Tfr *g.Om0
  vlam = g.vhw *(2*St +St**2)/(1+St**2)

  vgas = vran +vlam
  return vgas


def dsigP_dt (sigP,sigF,qP,TcolP,hP,hF):
  """
  Calculates the planetesimal fragmentation rate. The total rate is the sum of
  the planetesimal-planetesimal collisions (PP) and planetesimal-fragment (PF)
  erosive conditions.  The latter rate is set by the doER flag.
  """
  #the contribution due to P-P collisions and P-F collisions
  PP=  -qP/(1+qP) *sigP/TcolP
  if g.doER: 
    hX = max(hP,hF)
    PF = -qP *sigF/TcolP *(hP/hX)
  else: PF=0
  return PP+PF


def get_Omega(a0, Mstar):
  """
  returns the local orbital frequency
  """
  Om = sqrt(g.Gn*Mstar/a0**3)
  return Om


def get_preTts (alpha_ss, a0, Hg, sigG, Mstar, Om0):
  preTts = 5e-3/alpha_ss *( Hg *a0 *sigG /Mstar)**-2 /Om0
  return preTts


def get_cs (Temp, mu=2.34):
  """
  returns the isothermal sound speed
  """
  cs = sqrt(g.kB*Temp /(mu*g.amu))#isothermal sound speed
  return cs


def calc_Tau (s, cs, rhog, vgas):
  """
  Calculates the friction time
  """
  lmfp = l_mfp(rhog)
  vth = (8/pi)**0.5 *cs#conversion to thermal speed
  nu_mol = 0.5*vth*lmfp
  Re_p = 2*s*vgas/nu_mol
  Cd = C_D(Re_p)
  if 4*s>9*lmfp:#Continuum
    Tfr = 8*s*g.rhos /(3*Cd *rhog *vgas)
  else:#Epstein
    Tfr = Tfr_Ep(s, rhog, vth, g.rhos)
  return Tfr


def eh_min (sp, vh):
  """
  Computes the mimimum Hill eccentricity planetesimals should have due to
  self-stirring.  May be important for large planetesimals
  """
  return v_esc(sp, g.rhos)/vh


def eh_max (a0, rho, Mstar):
  """
  Computes the maximum Hill eccentricity corresponding to the escape
  speed of the embryos
  """
  al = alpha_a (a0, rho=1.0, Mstar=g.Msun)
  return sqrt(6/al)


def h_eff(tau, inc, Hg):
  """
  Computes the effective scaleheight.  Here we take account of:
    - turbulent diffusion
    - Keplerian inclination

  When the doTD switch is set, the maximum of the two is taken for the
  effective height
  """
  if g.doTD:
    heff = max(Hg*sqrt(g.alSS/(g.alSS+tau)), inc*g.a0)
  else:
    heff = inc*g.a0
  return heff


def inc_val (ecc, s):
  """
  ecc = (orbital) eccentricity
  s = planetesimal size
  """
  if g.incMode=='eq': inc = ecc/2
  elif g.incMode=='pollack': inc = v_esc(s, g.rhos) /(g.a0* g.Om0*sqrt(3))
  elif g.incMode=='2d': inc = 0
  else:
    print "[rout.inc_val]Error: no valid inclination mode"
    sys.exit()
  return inc


def l_mfp (rhog):
  """
  The mean-free path according to Nakagawa et al. (1986)
  """
  return 2e-9/rhog#mean free path (gas)


def load_matPars (fname):
  f = open(fname+'.txt','r')
  lines = f.readlines()
  f.close
  pars={}
  for line in lines:
    if line[0]<>'#':
      b=string.split(line,':')
      pars[string.rstrip(b[0])] = string.rstrip(b[1])
  return pars


def loadvariables (fname='variables.txt'):
  f = open(g.coregrowthdir+'/doc/'+fname,'r')
  lines = f.readlines()
  f.close
  variables={}
  for line in lines:
    if line[0]<>'#':
      b=string.split(line,':')
      variables[string.rstrip(b[0])] = string.rstrip(b[1])

  return variables


def mass2rad (mass, rho=1.0):
  """
  returns the radius, given the mass
  """
  return (0.75*mass/pi/rho)**(1./3)


def mmsn (str_out, a=1, p=-1.5, q=-0.5, fsig=1.0, fsigd=1.0, Mtot=0.013*g.Msun, M_star=g.Msun, 
         a_in=0, a_out=30, a_snow=2.7):
  """
  a set of functions that give default standard parameters corresponding to an MMSN disk
  Numerical values are from Nakagawa et al. (1996)

  Input parameters:
    a     :: semi major axis in AU
    p     :: surface density profile (assumed to be -1.5)
    q     :: temperature profile exponent (assumed to be -0.5)
    Mtot  :: total gas mass (assumed to be 1% of the solar mass)
    fsig  :: enhancement factor of surface density (i.e., Mtot) over 
             the canoncial (enhances solid & gas component)
    fsigd :: enhances dust:gas ratio (affects only the solid component)

  Output parameters:
    Sig   :: Sigma gas
    Sig_d :: Sigma solids
    f_gd  :: gas:dust ratio by mmass
    cs    :: isothermal sound speed
    rhog  :: gas density
    lmfp  :: mean free path
  """

  #First calculate the gas surface density
  #at 1 AU
  Sig1au = fsig*Mtot*(2+p) /(2*pi*g.au**2) /(a_out**(2+p) -a_in**(2+p))
  Sig_gas = Sig1au *a**p

  #then the dust sigma
  if a>a_snow:
    f_gd = 56.7/fsigd
  else:
    f_gd = 239./fsigd
  Sig = Sig_gas/f_gd


  Temp = 280 *a**q
  cs = get_cs (Temp)
  Om = get_Omega (a*g.au, g.Mstar)
  Hg = cs/Om
  rhog = Sig_gas/Hg/sqrt(2*pi)#@midplane
  lmfp = l_mfp(rhog)

  strarr = ['Sig', 'Sig_d', 'f_gd', 'cs', 'Om', 'Hg', 'rhog', 'lmfp', 'Temp']
  valarr = [Sig_gas, Sig, f_gd, cs, Om, Hg, rhog, lmfp, Temp]
  outval = []
  for str in str_out:
    c=strarr.count(str)
    if c>0:
      i=strarr.index(str)
      outval.append(valarr[i])
    else:
      outval.append(-1)

  return outval


def print_iniPars (svararr='', vararr=''):
  #TBD: maybe clean this up a bit?
  """
  Prints the initial variables
  """
  if svararr=='':#load global variables
    f = open(g.coregrowthdir+'/doc/parameters.txt','r')
    lines = f.readlines()
    f.close
    Nvars = len(lines)
  else:
    Nvars = len(svararr)

  for i in range(Nvars):
    if svararr<>'':
      name = svararr[i]
      var = vararr[i]
      doShow = True
    else:
      if lines[i][0]<>'#':
        b=string.split(lines[i],':')
        name = string.rstrip(b[0])
        desc = string.rstrip(b[1])
        var = eval('g.'+name)
        doShow = True
      else:
        doShow = False
    if doShow:
      sname = '%10s'%name+' = '
      if type(var) is bool:
        print sname, var
      elif type(var) is int:
        fmt='%8i'
        svar = fmt%var
        print sname, svar
      elif type(var) is float or type(var) is float64:
        fmt='%8.2e'
        svar = fmt%var
        print sname, svar
          

def rad2mass (rad, rho=1.0):
  """
  returns the mass, given the radius
  """
  return 4*pi/3 *rho *rad**3


def readData (sQarr, fname='data.dat', dtype='Float'):
  f = open(fname, 'r')
  hdrline = f.readline()
  f.close()
  sentr = hdrline.split()[1:]
  idx = [sentr.index(s) for s in sQarr]

  if dtype=='Float': data = loadtxt(fname, usecols=idx)
  elif dtype=='Str': data = loadtxt(fname, usecols=idx, dtype='str')
  return data


def sF_ini():
  """
  Provides the initial fragment size based on the minimum of the Qd
  curve which is approximately the point where the Qs and Qg terms
  meet, so we equate

    Qd_s(s) = Qd_g(s)

  and solve for s
  """
  if g.Q0s==0: 
    sFini=g.s_min
  else:
    sFini = (g.Q0g*g.rhos/g.Q0s)**(1./(g.bs-g.bg))
    if sFini<g.s_min: sFini = g.s_min

  return sFini


def sigma_E (Mc, Rh, a0, btilde):
  """
  Return the surface density in embryos, 
  corresponding to the protoplanet mass Mp
  """
  NcolE = (2*pi*a0 *Rh*btilde)**-1
  sigE = Mc*NcolE
   
  return sigE


def Tfr_Ep (s,rho_g,vth,rho_p=3.0):
  """
  Friction time in the Epstein limit
  """
  return rho_p*s/(vth*rho_g)


def v_drift(tau, vran):
  """
  Radial drift velocity

  Note the inclusion of the systematic term
  """
  vKep = g.a0 *g.Om0
  vdrift = tau/(1+tau**2) *(2*g.vhw +vran**2/vKep)
  return vdrift


def v_esc (rad, rhos):
  """
  Escape velocity
  """
  mass = rad2mass(rad, rho=rhos)
  return sqrt(2*g.Gn*mass/rad)


def writeData (iiter, sQarr, Qarr, fname='data.dat'):
  """
  write/append data to file fname
  """
  if iiter==0: 
    f=open(fname, 'w')
    line = '#entries:  irun  '
    for s in sQarr: line = line+s+'  '
    f.write(line+'\n')
  else: 
    f=open(fname, 'a')
  line = '%8d'%iiter
  fmt = '%10.2e'
  for F in Qarr:
    if isinstance(F, str): line = line+'  '+F
    else:
      strF = fmt%F
      line = line+strF
  line = line+'\n'
  f.write(line)
  f.close()

